
// Author:  David Martin  <dmartin@eecs.berkeley.edu>

import java.util.*;
import java.io.*;
import java.net.*;
import gnu.regexp.*;

class Segmentation
{
    public int width;
    public int height;
    public String image;
    public String user;
    public boolean gray;
    public boolean invert;
    public boolean flipflop;

    short[] mask;
    Util.IntArray pixelsPerSegment;	

    // Create segmentation will all pixels initially in 
    // segment segID.
    public Segmentation (int width, int height, int segID,
			 String image, String user, 
			 boolean gray, boolean invert, boolean flipflop) {
	Util.assert (segID >= 0);
	this.width = width;
	this.height = height;
	this.mask = new short [height * width];
	if (segID != 0) {
	    for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
		    mask[y*width+x] = (short) segID;
		}
	    }
	}
	pixelsPerSegment = new Util.IntArray ();
	pixelsPerSegment.set (segID, width * height);
	this.image = image;
	this.user = user;
	this.gray = gray;
	this.invert = invert;
	this.flipflop = flipflop;
    }

    // Parse segmentation file.
    public Segmentation (BufferedReader reader) throws Exception {
	String line;
	RE re;
	REMatch rem;
	int lineNo = 0;

	// Parse the header.
	width = -1;
	height = -1;
	user = null;
	image = null;
	int segments = -1;
	while (true) {
	    line = reader.readLine ();
	    lineNo++;
	    if (line == null) {
		throw new Exception ("Premature EOF at line " + lineNo 
				     + " while parsing segmentation.");
	    }

	    // Clean up line.
	    line = new RE ("\\s+").substituteAll (line, " ");	// Compress whitespace.
	    line = new RE ("#.*$").substitute (line, "");	// Strip comments.
	    line = new RE ("^ ").substitute (line, "");		// Remove leading whitespace.
	    line = new RE (" $").substitute (line, "");		// Remove trailing whitespace.
	    if (line.equals ("")) { continue; }			// Skip blank lines.

	    // 'data' marks end of header / start of data.
	    if (line.equals ("data")) {
		break;
	    }

	    re = new RE ("^width (\\d+)$");	
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		width = Integer.parseInt (rem.toString (1));
		if (width < 1) { 
		    throw new Exception (
			"Illegal width (" + width + ") at line " + lineNo + "."); 
		}
		continue;
	    }
	    re = new RE ("^height (\\d+)$");	
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		height = Integer.parseInt (rem.toString (1));
		if (height < 1) { 
		    throw new Exception (
			"Illegal height (" + height + ") at line " + lineNo + ".");
		}
		continue;
	    }
	    re = new RE ("^segments (\\d+)$");	
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		segments = Integer.parseInt (rem.toString (1));
		if (segments < 1) { 
		    throw new Exception (
			"Illegal segment count (" + segments + ") at line " + lineNo + ".");
		}
		continue;
	    }
	    re = new RE ("^image (.+)$");	
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		image = rem.toString (1);
		continue;
	    }
	    re = new RE ("^user (.+)$");	
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		user = rem.toString (1);
		continue;
	    }
	    re = new RE ("^gray (0|1)$");
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		gray = (Integer.parseInt (rem.toString (1)) == 1);
		continue;
	    }
	    re = new RE ("^invert (0|1)$");
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		invert = (Integer.parseInt (rem.toString (1)) == 1);
		continue;
	    }
	    re = new RE ("^flipflop (0|1)$");
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		flipflop = (Integer.parseInt (rem.toString (1)) == 1);
		continue;
	    }
	    re = new RE ("^format (ascii|binary) (cr|map)$");
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		String format = rem.toString(1);
		if (format.compareTo("ascii") != 0) {
		    throw new Exception (
			"format " + format + " not supported.  Must be ascii.");
		}
		format = rem.toString(2);
		if (format.compareTo("cr") != 0) {
		    throw new Exception (
			"format " + format + " not supported.  Must be cr (compressed-row).");
		}
		continue;
	    }
	    re = new RE ("^date (.*)$"); 
	    if (re.isMatch (line)) {
		continue;
	    }
	    throw new Exception (
		"Unknown token at line " + lineNo + ": " + line);
	}

	// Make sure we parsed valid width/height/segments values.
	if (width == -1) {
	    throw new Exception (
		"Width not found in file header.");
	}
	if (height == -1) {
	    throw new Exception (
		"Height not found in file header.");
	}
	if (segments == -1) {
	    throw new Exception (
		"Segment count not found in file header.");
	}
	
	// Parse the data.
	mask = new short [height * width];
	pixelsPerSegment = new Util.IntArray ();
	for (int y = 0; y < height; y++) {
	    for (int x = 0; x < width; x++) {
		mask[y*width+x] = (short) segments;
	    }
	}
	re = new RE ("^(\\d+) (\\d+) (\\d+) (\\d+)$");
	while (true) {
	    line = reader.readLine ();
	    lineNo++;
	    if (line == null) { break; }
	    if (!re.isMatch (line)) {
		throw new Exception (
		    "Parse error at line " + lineNo + ": " + line);
	    }
	    rem = re.getMatch (line);
	    int seg = Integer.parseInt (rem.toString (1));
	    int row = Integer.parseInt (rem.toString (2));
	    int col1 = Integer.parseInt (rem.toString (3));
	    int col2 = Integer.parseInt (rem.toString (4));
	    if (seg < 0 || seg >= segments) {
		throw new Exception (
		    "Illegal segment ID (" + seg + ") at line " + lineNo + ".");
	    }
	    if (row < 0 || row >= height) {
		throw new Exception (
		    "Illegal row (" + row + ") at line " + lineNo + ".");
	    }
	    if (col1 < 0 || col1 >= width) {
		throw  new Exception (
		    "Illegal start column (" + col1 + ") at line " + lineNo + ".");
	    }
	    if (col2 < 0 || col2 >= width) {
		throw  new Exception (
		    "Illegal end column (" + col2 + ") at line " + lineNo + ".");
	    }
	    // Set the relevant mask values.
	    for (int col = col1; col <= col2; col++) {
		if (mask[row*width+col] != segments) {
		    throw new Exception (
			"Pixel (" + col + "," + row + ") is claimed by "
			+ "both segment " + mask[row*width+col]
			+ " and segment + " + seg + ".");
		}
		mask[row*width+col] = (short) seg;
		pixelsPerSegment.set (seg, pixelsPerSegment.get (seg) + 1);
	    }
	}

	// Make sure the mask is complete.
	for (int y = 0; y < height; y++) {
	    for (int x = 0; x < width; x++) {
		if (mask[y*width+x] == segments) {
		    throw new Exception (
			"Pixel (" + x + "," + y + ") was not found.");
		}
	    }
	}
		
	// Make sure the segment IDs are dense.
	TreeSet set = new TreeSet ();
	for (int y = 0; y < height; y++) {
	    for (int x = 0; x < width; x++) {
		set.add (new Integer (mask[y*width+x]));
	    }
	}
	Iterator iter = set.iterator();
	int seg = 0;
	while (iter.hasNext()) {
	    Integer val = (Integer) iter.next ();
	    if (val.intValue() != seg) {
		throw new Exception (
		    "Segment ID " + seg + "is missing from the file.");
	    }
	    seg++;
	}
    }

    public static Segmentation openSegmentationFile (File file) throws Exception {
	FileReader reader = null;
	try {
	    reader = new FileReader (file);
	    return new Segmentation (new BufferedReader (reader));
	} finally {
	    try {
		if (reader != null) { reader.close (); }
	    } catch (Exception e) {}
	}
    }

    public static Segmentation openSegmentationURL (URL url) throws Exception {
	InputStream strm = null;
	try {
	    strm = url.openStream ();
	    Reader reader = new InputStreamReader (strm);
	    return new Segmentation (new BufferedReader (reader));
	} finally {
	    try {
		if (strm != null) { strm.close (); }
	    } catch (Exception e) {}
	}
    }

    public int getImageID () throws Exception {
	return Integer.parseInt (image);
    }

    public int getUserID () throws Exception {
	return Integer.parseInt (user); 
    }

    public Reader getReader () throws Exception {
	PipedReader reader = new PipedReader ();
	PipedWriter writer = new PipedWriter (reader);
	SegmentationGenerator gen = new SegmentationGenerator (writer);
	gen.start ();
	return reader;
    }

    public int pixelCount (int segID) {
	return pixelsPerSegment.get (segID);
    }

    public int nonEmptySegmentCount () {
	int count = 0;
	for (int i = 0; i < pixelsPerSegment.length(); i++) {
	    if (pixelsPerSegment.get(i) > 0) {
		count++;
	    }
	}
	return count;
    }

    public int[] nonEmptySegmentIDs () {
	int[] ids = new int [nonEmptySegmentCount()];
	int c = 0;
	for (int i = 0; i < pixelsPerSegment.length(); i++) {
	    if (pixelsPerSegment.get(i) > 0) {
		ids[c++] = i;
	    }
	}
	return ids;
    }

    public int get (int x, int y) {
	return mask[y*width+x];
    }

    public void set (int x, int y, int val) {
	pixelsPerSegment.set (mask[y*width+x], pixelsPerSegment.get (mask[y*width+x]) - 1);
	mask[y*width+x] = (short) val;
	pixelsPerSegment.set (mask[y*width+x], pixelsPerSegment.get (mask[y*width+x]) + 1);
    }

    public int newSegmentID () {
	for (int id = 0; ; id++) {
	    if (pixelsPerSegment.get (id) == 0) {
		return id;
	    }
	}
    }

    // Flipflop the data.
    public void doFlipFlop () {
	int n = width * height;
	for (int i = 0, j = n - 1; i < j; i++, j--) {
	    short tmp = mask[i];
	    mask[i] = mask[j];
	    mask[j] = tmp;
	}	
    }

    // Create a thread that writes a segmentation file to the given writer.
    class SegmentationGenerator 
	extends Thread 
    {
	Writer writer;

	public SegmentationGenerator (Writer writer) {
	    this.writer = writer;
	}

	void writeSegFile () throws Exception {
	    // Remap segment IDs so that the non-empty segments are
	    // densly packed starting at zero.
	    int[] ids = nonEmptySegmentIDs ();		// Get sparse id list.
	    int numSegments = ids.length;
	    int maxID = 0;				// Find max id.
	    for (int i = 0; i < ids.length; i++) {	
		maxID = Math.max (maxID, ids[i]);
	    }
	    int[] idMap = new int [maxID + 1];		// Init idMap.
	    for (int i = 0; i < idMap.length; i++) {	
		idMap[i] = -1;
	    }
	    for (int i = 0; i < ids.length; i++) {	// Map non empty ids.
		idMap[ids[i]] = i;
	    }

	    if (image != null) {
		writer.write ("image " + image + "\n");
	    }
	    if (user != null) {
		writer.write ("user " + user + "\n");
	    }
	    writer.write ("width " + width + "\n");
	    writer.write ("height " + height + "\n");
	    writer.write ("segments " + numSegments + "\n");
	    writer.write ("gray " + (gray ? 1 : 0) + "\n");
	    writer.write ("invert " + (invert ? 1 : 0) + "\n");
	    writer.write ("flipflop " + (flipflop ? 1 : 0) + "\n");
	    writer.write ("format ascii cr\n");
	    writer.write ("data\n"); // Marks end of header / beginning of data.
	    // Each line contains: <s> <r> <c1> <c2>
	    // This means that pixels [c1..c2] of row r belong to segment s.
	    for (int row = 0; row < height; row++) {
		int prevSeg = -1;
		for (int col = 0; col < width; col++) {
		    int seg = get (col, row);
		    if (seg != prevSeg) {
			if (col > 0) {
			    int a = col - 1;
			    writer.write (a + "\n");
			}
			int index = idMap[seg];
			Util.assert (index >= 0 && index < numSegments);
			writer.write (index + " " + row + " " + col + " ");
		    }
		    prevSeg = seg;
		}
		int a = width - 1;
		writer.write (a + "\n");
	    }
	}

	public void run () {
	    try {
		writeSegFile ();
	    } catch (Exception e) {
		// Not sure what to do about an exception here.
	    } finally {
		try { writer.close (); } catch (Exception e) {}
	    }
	}
    }
}
